# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2022 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Community PID slug field."""

from uuid import UUID

from invenio_db import db
from invenio_pidstore.errors import PIDDeletedError, PIDDoesNotExistError
from invenio_records.systemfields import SystemField, SystemFieldContext
from invenio_records_resources.records.api import PersistentIdentifierWrapper


class PIDSlugFieldContext(SystemFieldContext):
    """PID Slug Field Context."""

    def parse_pid(self, value):
        """Parse pid."""
        if isinstance(value, UUID):
            return value
        try:
            return UUID(value)
        except (TypeError, ValueError):
            return value

    def resolve(self, pid_value, registered_only=True):
        """Resolve identifier (either uuid or slug)."""
        pid_value = self.parse_pid(pid_value)

        if isinstance(pid_value, UUID):
            field_name = self.field._id_field
        else:
            field_name = self.field._slug_field
            # Edge case since we allow nullable slug in the database table.
            if not pid_value:
                raise PIDDoesNotExistError("comid", "")

        with db.session.no_autoflush:  # avoid flushing the current session
            model = self.record_cls.model_cls.query.filter_by(
                **{field_name: pid_value}
            ).one_or_none()
            if model is None:
                raise PIDDoesNotExistError("comid", str(pid_value))
            record = self.record_cls(model.data, model=model)
            if record.is_deleted:
                raise PIDDeletedError(PersistentIdentifierWrapper(pid_value), record)
            return record


class PIDSlugField(SystemField):
    """System field for managing record access."""

    def __init__(self, id_field, slug_field):
        """Create a new RecordAccessField instance."""
        self._id_field = id_field
        self._slug_field = slug_field

    def obj(self, record):
        """Get the access object."""
        pid_value = getattr(record, self._id_field)
        if pid_value is None:
            return None
        return PersistentIdentifierWrapper(str(pid_value))

    def __get__(self, record, owner=None):
        """Get the record's access object."""
        if record is None:
            # access by class
            return PIDSlugFieldContext(self, owner)

        # access by object
        return self.obj(record)

    #
    # Life-cycle hooks
    #
    def pre_commit(self, record):
        """Called before a record is committed."""
        # Make sure we don't dump the two model fields into the JSON of the
        # record.
        record.pop(self._id_field, None)
        record.pop(self._slug_field, None)



#
# This file is part of Invenio.
# Copyright (C) 2023 TU Wien.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Add deletion status to communities."""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "5c68d45c80f0"
down_revision = "02cd82910727"
branch_labels = ()
depends_on = None


def _get_default_deletion_status():
    """Try to get the default value for record deletion status."""
    try:
        # first try: get the default value from the DB model class
        from invenio_communities.communities.records.models import CommunityMetadata

        default_value = CommunityMetadata.deletion_status.default.arg
        assert len(default_value) == 1

    except Exception:
        try:
            # second try: try to get the value from the enum (more flaky)
            from invenio_communities.communities.records.systemfields.deletion_status import (  # noqa
                CommunityDeletionStatusEnum,
            )

            default_value = CommunityDeletionStatusEnum.PUBLISHED.value
            assert len(default_value) == 1

        except Exception:
            # fallback: just use 'P', as that's the current default (most flaky)
            default_value = "P"

    return default_value


def upgrade():
    """Upgrade database."""
    # step 1: create the columns, but make them nullable for now
    op.add_column(
        "communities_metadata",
        sa.Column("deletion_status", sa.String(length=1), nullable=True),
    )

    # step 2: set default values for existing rows
    default_value = _get_default_deletion_status()
    metadata_table = sa.sql.table(
        "communities_metadata", sa.sql.column("deletion_status")
    )
    op.execute(metadata_table.update().values(deletion_status=default_value))

    # step 3: make the original table not nullable
    op.alter_column("communities_metadata", "deletion_status", nullable=False)


def downgrade():
    """Downgrade database."""
    op.drop_column("communities_metadata", "deletion_status")



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 Northwestern University.
# Copyright (C) 2022-2024 CERN.
# Copyright (C) 2022-2023 Graz University of Technology.
#
# Invenio-Communities is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Members service."""

from datetime import datetime, timezone

from flask import current_app
from invenio_access.permissions import system_identity
from invenio_accounts.models import Role
from invenio_i18n import gettext as _
from invenio_notifications.services.uow import NotificationOp
from invenio_records_resources.services import LinksTemplate
from invenio_records_resources.services.records import (
    RecordService,
    ServiceSchemaWrapper,
)
from invenio_records_resources.services.uow import (
    IndexRefreshOp,
    RecordCommitOp,
    RecordDeleteOp,
    unit_of_work,
)
from invenio_requests import current_events_service, current_requests_service
from invenio_requests.customizations.event_types import CommentEventType
from invenio_search.engine import dsl
from kombu import Queue
from marshmallow import ValidationError
from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm.exc import NoResultFound
from werkzeug.local import LocalProxy

from ...notifications.builders import CommunityInvitationSubmittedNotificationBuilder
from ...proxies import current_roles
from ..errors import AlreadyMemberError, InvalidMemberError
from ..records.api import ArchivedInvitation
from .request import CommunityInvitation, MembershipRequestRequestType
from .schemas import (
    AddBulkSchema,
    DeleteBulkSchema,
    InvitationDumpSchema,
    InviteBulkSchema,
    MemberDumpSchema,
    PublicDumpSchema,
    RequestMembershipSchema,
    UpdateBulkSchema,
)


def invite_expires_at():
    """Get the invitation expiration date."""
    return (
        datetime.utcnow().replace(tzinfo=timezone.utc)
        + current_app.config["COMMUNITIES_INVITATIONS_EXPIRES_IN"]
    )


class MemberService(RecordService):
    """Community members service."""

    @property
    def community_cls(self):
        """Return community class."""
        return self.config.community_cls

    @property
    def member_dump_schema(self):
        """Schema for creation."""
        return ServiceSchemaWrapper(self, schema=MemberDumpSchema)

    @property
    def public_dump_schema(self):
        """Schema for creation."""
        return ServiceSchemaWrapper(self, schema=PublicDumpSchema)

    @property
    def invitation_dump_schema(self):
        """Schema for creation."""
        return ServiceSchemaWrapper(self, schema=InvitationDumpSchema)

    @property
    def add_schema(self):
        """Schema for creation."""
        return ServiceSchemaWrapper(self, schema=AddBulkSchema)

    @property
    def invite_schema(self):
        """Schema for creation."""
        return ServiceSchemaWrapper(self, schema=InviteBulkSchema)

    @property
    def update_schema(self):
        """Schema for bulk update."""
        return ServiceSchemaWrapper(self, schema=UpdateBulkSchema)

    @property
    def delete_schema(self):
        """Schema for bulk delete."""
        return ServiceSchemaWrapper(self, schema=DeleteBulkSchema)

    @property
    def request_membership_schema(self):
        """Wrapped schema for request membership."""
        return ServiceSchemaWrapper(self, schema=RequestMembershipSchema)

    @property
    def archive_indexer(self):
        """Factory for creating an indexer instance."""
        return self.config.archive_indexer_cls(
            # the routing key is mandatory in the indexer constructor since
            # it is afterwards passed explicitly to the created consumers
            # and producers. this means that it is not strictly necessary to
            # pass it to the queue constructor. however, it is passed for
            # consistency (in case the queue is used by itself) and to help
            # entity declaration on publish.
            queue=LocalProxy(
                lambda: Queue(
                    self.config.archive_indexer_queue_name,
                    exchange=current_app.config["INDEXER_MQ_EXCHANGE"],
                    routing_key=self.config.archive_indexer_queue_name,
                )
            ),
            routing_key=self.config.archive_indexer_queue_name,
            record_cls=self.config.archive_cls,
            record_to_index=self.record_to_index,
            record_dumper=self.config.index_dumper,
        )

    #
    # Add and invite
    #
    @unit_of_work()
    def add(self, identity, community_id, data, uow=None):
        """Add group members.

        The default permission policy only allow groups to be added. Users must
        be invited.
        """
        ret = self._create(
            identity,
            community_id,
            data,
            self.add_schema,
            "members_add",
            self._add_factory,
            uow,
        )
        # ensure index is refreshed to search for newly added members
        uow.register(IndexRefreshOp(indexer=self.indexer))
        return ret

    @unit_of_work()
    def invite(self, identity, community_id, data, uow=None):
        """Invite group members.

        Only users and email member types can be invited, and a member can only
        have one invitation per community

        Email member type is not yet supported.
        """
        ret = self._create(
            identity,
            community_id,
            data,
            self.invite_schema,
            "members_invite",
            self._invite_factory,
            uow,
        )

        # ensure index is refreshed to search for newly added members
        uow.register(IndexRefreshOp(indexer=self.indexer))
        return ret

    def _create(self, identity, community_id, data, schema, action, factory, uow):
        """Internal method used for both adding and inviting users/groups.

        The methods are shared because adding and inviting users/groups share
        the most of the same permission and integrity checks.
        """
        community = self.community_cls.get_record(community_id)
        # Validate data (if there are errors, .load() raises)
        data, errors = schema.load(
            data,
            context={"identity": identity},
        )
        role = data["role"]
        members = data["members"]
        member_types = {m["type"] for m in data["members"]}
        message = data.get("message")
        # Users are expected to explicitly change their visibility themselves
        # due to data privacy concerns.
        # The system identity can set the visible property to support data
        # migration use cases.
        visible = data.get("visible", False)
        if visible and identity != system_identity:
            raise ValidationError(_("Must be false"))

        # Permission checks - validates that:
        # - identity has permission to manage members
        # - identity has permission to add/invite for a given role (e.g.
        #   managers should not be able to invite owners because it would make
        #   privilege escalation possible).
        # - identity is allowed to create a specific member type (e.g. usually
        #   we only allow adding groups, and inviting users/emails however in
        #   a migration use case we may want to allow adding users directly).
        self.require_permission(
            identity,
            action,
            record=community,
            role=role.name,
            member_types=member_types,
        )

        # Add/invite members via the factory function.
        for m in members:
            # TODO: Add support for inviting an email
            if m["type"] == "email":
                raise ValidationError(_("Invalid member type: email"))

            factory(identity, community, role, visible, m, message, uow)
            # Run components
            self.run_components(
                action,
                identity,
                record=m,
                community=community,
                errors=None,
                uow=uow,
            )

        return True

    def _add_factory(
        self,
        identity,
        community,
        role,
        visible,
        member,
        message,
        uow,
        active=True,
        request_id=None,
    ):
        """Add a member to the community."""
        member_arg = {member["type"] + "_id": member["id"]}
        try:
            # Integrity checks happens here which will validate:
            # - if a user/group is already a member
            # - if a user is already invited
            member = self.record_cls.create(
                {},
                community_id=community.id,
                role=role.name,
                active=active,
                visible=visible,
                request_id=request_id,
                **member_arg,
            )
        except IntegrityError as e:
            raise AlreadyMemberError() from e

        # TODO: Change this to bulk index all member records in one request
        # instead of N single indexing requests.
        uow.register(RecordCommitOp(member, indexer=self.indexer))

    def _invite_factory(self, identity, community, role, visible, member, message, uow):
        """Invite a member to the community."""
        if member["type"] == "group":
            # Groups cannot be invited, because groups have no one who can
            # accept an invitation.
            raise InvalidMemberError(member)

        # Add member entry
        if member["type"] == "user":
            # Create request
            title = _('Invitation to join "{community}"').format(
                community=community.metadata["title"],
            )
            description = _('You will join as "{role}".').format(role=role.title)

            request_item = current_requests_service.create(
                identity,
                {
                    "title": title,
                    "description": description,
                },
                CommunityInvitation,
                {"user": member["id"]},
                creator=community,
                # TODO: perhaps topic should be the actual membership record
                # instead
                topic=community,
                expires_at=invite_expires_at(),
                uow=uow,
            )

            # message was provided.
            if message:
                data = {"payload": {"content": message}}
                current_events_service.create(
                    identity,
                    request_item.id,
                    data,
                    CommentEventType,
                    uow=uow,
                    notify=False,
                )

            uow.register(
                NotificationOp(
                    CommunityInvitationSubmittedNotificationBuilder.build(
                        request=request_item._request,
                        # explicit string conversion to get the value of LazyText
                        role=str(role.title),
                        message=message,
                    )
                )
            )

            # Create an inactive member entry linked to the request.
            self._add_factory(
                identity,
                community,
                role,
                visible,
                member,
                message,
                uow,
                active=False,
                request_id=request_item.id,
            )

    def search(
        self,
        identity,
        community_id,
        params=None,
        search_preference=None,
        extra_filter=None,
        **kwargs,
    ):
        """Search."""
        # Apply extra filters
        filter_ = dsl.Q("term", **{"active": True})
        if extra_filter:
            filter_ &= extra_filter

        return self._members_search(
            identity,
            community_id,
            "members_search",
            self.member_dump_schema,
            self.config.search,
            extra_filter=filter_,
            params=params,
            search_preference=search_preference,
            **kwargs,
        )

    def scan(
        self,
        identity,
        community_id,
        params=None,
        search_preference=None,
        extra_filter=None,
        **kwargs,
    ):
        """Scan community members to retrieve all matching the query."""
        # Apply extra filters
        filter_ = dsl.Q("term", **{"active": True})
        if extra_filter:
            filter_ &= extra_filter

        return self._members_search(
            identity,
            community_id,
            "members_search",
            self.member_dump_schema,
            self.config.search,
            extra_filter=filter_,
            scan_params=params,
            search_preference=search_preference,
            scan=True,
            **kwargs,
        )

    def search_public(
        self, identity, community_id, params=None, search_preference=None, **kwargs
    ):
        """Search public members matching the querystring."""
        # The search for members is split two methods (public, members) to
        # prevent leaking of information. E.g. the public serialization
        # must not have all fields present.
        # TODO: limit fields on which the query works on to avoid leaking information
        return self._members_search(
            identity,
            community_id,
            "members_search_public",
            self.public_dump_schema,
            self.config.search_public,
            extra_filter=(
                dsl.Q("term", **{"visible": True}) & dsl.Q("term", **{"active": True})
            ),
            params=params,
            search_preference=search_preference,
            **kwargs,
        )

    def search_invitations(
        self, identity, community_id, params=None, search_preference=None, **kwargs
    ):
        """Search invitations."""
        # The search for invitations used the ArchivedInvitation record class
        # which will search over the "communitymembers" alias which include
        # both current invitations and archived invitations.
        return self._members_search(
            identity,
            community_id,
            "search_invites",
            self.invitation_dump_schema,
            self.config.search_invitations,
            record_cls=ArchivedInvitation,
            extra_filter=dsl.Q("term", **{"active": False}),
            params=params,
            search_preference=search_preference,
            **kwargs,
        )

    def _members_search(
        self,
        identity,
        community_id,
        permission_action,
        schema,
        search_opts,
        extra_filter=None,
        params=None,
        search_preference=None,
        scan=False,
        scan_params=None,
        **kwargs,
    ):
        """Members search."""
        community = self.community_cls.get_record(community_id)
        self.require_permission(identity, permission_action, record=community)

        # Apply extra filters
        filter = dsl.Q("term", **{"community_id": community.id})
        if extra_filter:
            filter &= extra_filter

        # Prepare and execute the search
        params = params or {}
        scan_params = scan_params or {}

        search = self._search(
            "search_members",
            identity,
            params,
            search_preference,
            search_opts=search_opts,
            permission_action=None,
            extra_filter=filter,
            **kwargs,
        )
        # scan has a default scroll timeout of 5 minutes
        # https://github.com/opensearch-project/opensearch-py/blob/fe3b5a8922aa8eb04f735c74d127d7ea68a00bec/opensearchpy/helpers/actions.py#L492-L503
        search_result = (
            search.params(**scan_params).scan() if scan else search.execute()
        )

        return self.result_list(
            self,
            identity,
            search_result,
            params,
            links_tpl=(
                None
                if scan
                else LinksTemplate(
                    self.config.links_search,
                    context={
                        "args": params,
                        "community_id": community_id,
                    },
                )
            ),
            schema=schema,
        )

    def read_memberships(self, identity):
        """Searches the memberships of a specific user/identity."""
        return {"memberships": self.config.record_cls.get_memberships(identity)}

    @unit_of_work()
    def update(self, identity, community_id, data, uow=None, refresh=False):
        """Bulk update.

        Used to update both active members and active invitations. Archived
        invitations cannot be updated.
        """
        community = self.community_cls.get_record(community_id)

        # Permission check - validates that:
        # - identity has permission to change any member at all (incl self)
        self.require_permission(
            identity,
            "members_bulk_update",
            record=community,
        )

        # Validate data (if there are errors, .load() raises)
        data, errors = self.update_schema.load(
            data,
            context={"identity": identity},
        )

        # Schema validates that role and/or visibility are defined.
        members = self.record_cls.get_members(community.id, members=data["members"])
        if len(members) != len(data["members"]):
            raise InvalidMemberError()
        role = data.get("role")
        visible = data.get("visible")

        # Perform updates (and check permissions)
        for m in members:
            self._update(identity, community, m, role, visible, uow)

        # Make sure we're not left owner-less if a role was changed.
        if role is not None:
            if not self.record_cls.has_members(
                community_id, role=current_roles.owner_role.name
            ):
                raise ValidationError(
                    _("A community must have at least one owner."),
                )

        if refresh:
            uow.register(IndexRefreshOp(indexer=self.indexer))

        return True

    def _update(self, identity, community, member, role, visible, uow):
        """Update a member's role/visibility."""
        # DO NOT USE DIRECTLY - always use update() which will correctly check
        # if we're left without an owner!
        self.require_permission(
            identity,
            "members_update",
            record=community,
            member=member,
            role=role.name if role is not None else None,
        )
        is_self = identity.id is not None and str(member.user_id) == str(identity.id)

        # Pre-conditions:
        # You cannot change your own role (owners, managers, members = all)
        # For owners/managers, this prevents them accidentally loosing access.
        # They will have to ask another owner/manager to change their role.
        # For curators/readers, they should not be be allowed to change their
        # own role. Having a business rule avoid making the permissions overly
        # complex.
        if role is not None and is_self:
            raise ValidationError(_("You cannot change your own role."))

        # Owners/managers can change visibility to false for users, and
        # true/false for groups. Users themselves can change their own
        # visibility. System identity can can always change for all.
        if visible is not None and member.user_id is not None:
            if visible and not (is_self or system_identity == identity):
                raise ValidationError(
                    _("You can only set public visibility on your own membership."),
                )

        # Update membership
        if role is not None:
            if not member.active:  # still an invitation
                data = {
                    "payload": {
                        "content": _(
                            'You will join as "{role}" (changed from: "{previous}").'
                        ).format(
                            role=role.title,
                            previous=member.role,
                        ),
                    }
                }
                current_events_service.create(
                    identity, member.request_id, data, CommentEventType, uow=uow
                )

            # update role
            member.role = role.name
        if visible is not None:
            member.visible = visible

        # Run components
        self.run_components(
            "members_update",
            identity,
            record=member,
            community=community,
            errors=None,
            uow=uow,
        )

        uow.register(RecordCommitOp(member, self.indexer))

        return True

    @unit_of_work()
    def delete(self, identity, community_id, data, uow=None):
        """Bulk delete."""
        community = self.community_cls.get_record(community_id)

        # Permission check - validates that:
        # - identity has permission to delete any member at all (incl self)
        self.require_permission(
            identity,
            "members_bulk_delete",
            record=community,
        )

        # Validate data (if there are errors, .load() raises)
        data, errors = self.delete_schema.load(
            data,
            context={"identity": identity},
        )
        members = self.record_cls.get_members(community.id, members=data["members"])
        if len(members) != len(data["members"]):
            raise InvalidMemberError()

        # Perform deletes (and check permissions)
        for m in members:
            self.require_permission(
                identity,
                "members_delete",
                record=community,
                member=m,
            )
            # Run components
            self.run_components(
                "members_delete",
                identity,
                record=m,
                community=community,
                errors=None,
                uow=uow,
            )
            uow.register(RecordDeleteOp(m, self.indexer, force=True))

        # Make sure we're not left owner-less
        if not self.record_cls.has_members(
            community_id, role=current_roles.owner_role.name
        ):
            raise ValidationError(
                _("A community must have at least one owner."),
            )

        return True

    @unit_of_work()
    def accept_invite(self, identity, request_id, uow=None):
        """Accept an invitation."""
        # Permissions are checked on the request action
        assert identity == system_identity
        member = self.record_cls.get_member_by_request(request_id)
        assert member.active is False
        archived_invitation = ArchivedInvitation.create_from_member(member)
        member.active = True
        # TODO: recompute permissions for member.
        # Run components
        self.run_components(
            "accept_invite",
            identity,
            record=member,
            errors=None,
            uow=uow,
        )

        uow.register(RecordCommitOp(member, indexer=self.indexer))
        uow.register(RecordCommitOp(archived_invitation, indexer=self.archive_indexer))
        uow.register(IndexRefreshOp(indexer=self.indexer))

    @unit_of_work()
    def decline_invite(self, identity, request_id, uow=None):
        """Decline an invitation."""
        # Permissions are checked on the request action
        assert identity == system_identity
        member = self.record_cls.get_member_by_request(request_id)
        assert member.active is False
        archived_invitation = ArchivedInvitation.create_from_member(member)
        uow.register(RecordDeleteOp(member, indexer=self.indexer, force=True))
        uow.register(RecordCommitOp(archived_invitation, indexer=self.archive_indexer))
        uow.register(
            IndexRefreshOp(
                # need to use an indexer with a diff index
                # no access to invitations indexer
                indexer=self.indexer,
                index=ArchivedInvitation.index,
            )
        )

    def read_many(self, *args, **kwargs):
        """Not implemented."""
        raise NotImplementedError("Use search() or search_public()")

    def read_all(self, *args, **kwargs):
        """Not implemented."""
        raise NotImplementedError("Use search() or search_public()")

    def read(self, *args, **kwargs):
        """Not implemented."""
        raise NotImplementedError("Use search() or search_public()")

    def create(self, *args, **kwargs):
        """Not implemented."""
        raise NotImplementedError("Use add() or invite()")

    def rebuild_index(self, identity, uow=None):
        """Reindex all records managed by this service.

        Note: Skips (soft) deleted records.
        """
        members = self.record_cls.model_cls.query.filter_by(is_deleted=False).all()
        self.indexer.bulk_index([member.id for member in members])

        archived_invitations = ArchivedInvitation.model_cls.query.filter_by(
            is_deleted=False
        ).all()
        self.archive_indexer.bulk_index([inv.id for inv in archived_invitations])

        return True

    # Request membership
    @unit_of_work()
    def request_membership(self, identity, community_id, data, uow=None):
        """Request membership to the community.

        A user can only have one request per community.

        All validations raise, so it's up to parent layer to handle them.
        """
        community = self.community_cls.get_record(community_id)

        data, errors = self.request_membership_schema.load(
            data,
            context={"identity": identity},
        )
        message = data.get("message", "")

        self.require_permission(
            identity,
            "request_membership",
            record=community,
        )

        # Create request
        title = _('Request to join "{community}"').format(
            community=community.metadata["title"],
        )
        request_item = current_requests_service.create(
            identity,
            data={
                "title": title,
                # "description": description,
            },
            request_type=MembershipRequestRequestType,
            receiver=community,
            creator={"user": str(identity.user.id)},
            topic=community,  # user instead?
            # TODO: Consider expiration
            # expires_at=invite_expires_at(),
            uow=uow,
        )

        if message:
            data = {"payload": {"content": message}}
            current_events_service.create(
                identity,
                request_item.id,
                data,
                CommentEventType,
                uow=uow,
                notify=False,
            )

        # TODO: Add notification mechanism
        # uow.register(
        #     NotificationOp(
        #         MembershipRequestSubmittedNotificationBuilder.build(
        #             request=request_item._request,
        #             # explicit string conversion to get the value of LazyText
        #             role=str(role.title),
        #             message=message,
        #         )
        #     )
        # )

        # Create an inactive member entry linked to the request.
        self._add_factory(
            identity,
            community=community,
            role=current_roles["reader"],
            visible=False,
            member={"type": "user", "id": str(identity.user.id)},
            message=message,
            uow=uow,
            active=False,
            request_id=request_item.id,
        )

        # No registered component with a request_membership method for now,
        # so no run_components for now.

        # Has to return the request so that frontend can redirect to it
        return request_item

    @unit_of_work()
    def update_membership_request(self, identity, community_id, data, uow=None):
        """Update membership request."""
        # TODO: Implement me
        pass

    def search_membership_requests(self):
        """Search membership requests."""
        # TODO: Implement me
        pass

    @unit_of_work()
    def accept_membership_request(self, identity, request_id, uow=None):
        """Accept membership request."""
        # TODO: Implement me
        pass

    @unit_of_work()
    def close_membership_request(self, identity, request_id, uow=None):
        """Close membership request.

        Used for cancelling, declining, or expiring a membership request.

        For now we just delete the "fake" member that was created in
        request_membership. TODO: explore alternatives/ramifications at a
        later point.
        """
        # Permissions are checked on the request action
        assert identity == system_identity
        member = self.record_cls.get_member_by_request(request_id)
        assert member.active is False
        uow.register(RecordDeleteOp(member, indexer=self.indexer, force=True))



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 Northwestern University.
# Copyright (C) 2022 CERN.
#
# Invenio-Communities is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Invenio Communities Resource API."""

from flask import g
from flask_resources import resource_requestctx, response_handler, route
from invenio_records_resources.resources.records.resource import (
    RecordResource,
    request_data,
    request_extra_args,
    request_search_args,
    request_view_args,
)
from invenio_records_resources.resources.records.utils import search_preference


class MemberResource(RecordResource):
    """Members resource."""

    def create_url_rules(self):
        """Create the URL rules for the record resource."""
        routes = self.config.routes
        return [
            route("POST", routes["members"], self.add),
            route("DELETE", routes["members"], self.delete),
            route("PUT", routes["members"], self.update),
            route("GET", routes["members"], self.search),
            route("GET", routes["publicmembers"], self.search_public),
            route("POST", routes["invitations"], self.invite),
            route("PUT", routes["invitations"], self.update_invitations),
            route("GET", routes["invitations"], self.search_invitations),
            route("POST", routes["membership_requests"], self.request_membership),
        ]

    @request_view_args
    @request_search_args
    @response_handler(many=True)
    def search(self):
        """Perform a search over the items."""
        hits = self.service.search(
            g.identity,
            resource_requestctx.view_args["pid_value"],
            params=resource_requestctx.args,
            search_preference=search_preference(),
        )
        return hits.to_dict(), 200

    @request_view_args
    @request_search_args
    @response_handler(many=True)
    def search_public(self):
        """Perform a search over the items."""
        hits = self.service.search_public(
            g.identity,
            resource_requestctx.view_args["pid_value"],
            params=resource_requestctx.args,
            search_preference=search_preference(),
        )
        return hits.to_dict(), 200

    @request_view_args
    @request_search_args
    @response_handler(many=True)
    def search_invitations(self):
        """Perform a search over the invitations."""
        hits = self.service.search_invitations(
            g.identity,
            resource_requestctx.view_args["pid_value"],
            params=resource_requestctx.args,
            search_preference=search_preference(),
        )
        return hits.to_dict(), 200

    @request_view_args
    @request_data
    def add(self):
        """Add members."""
        self.service.add(
            g.identity,
            resource_requestctx.view_args["pid_value"],
            resource_requestctx.data,
        )
        return "", 204

    @request_view_args
    @request_data
    def invite(self):
        """Invite members."""
        self.service.invite(
            g.identity,
            resource_requestctx.view_args["pid_value"],
            resource_requestctx.data,
        )
        return "", 204

    @request_view_args
    @request_data
    def request_membership(self):
        """Request membership."""
        request = self.service.request_membership(
            g.identity,
            resource_requestctx.view_args["pid_value"],
            resource_requestctx.data,
        )
        return request.to_dict(), 201

    @request_view_args
    @request_extra_args
    @request_data
    def update(self):
        """Update member."""
        self.service.update(
            g.identity,
            resource_requestctx.view_args["pid_value"],
            data=resource_requestctx.data,
            refresh=resource_requestctx.args.get("refresh", False),
        )
        return "", 204

    @request_view_args
    @request_data
    def update_invitations(self):
        """Update invitations."""
        self.service.update(
            g.identity,
            resource_requestctx.view_args["pid_value"],
            data=resource_requestctx.data,
        )
        return "", 204

    @request_view_args
    @request_data
    def delete(self):
        """Delete members."""
        self.service.delete(
            g.identity,
            resource_requestctx.view_args["pid_value"],
            data=resource_requestctx.data,
        )
        return "", 204



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023-2024 CERN.
# Copyright (C) 2023 TU Wien.
#
# Invenio-Communities is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.
"""Communities user moderation actions."""

from collections import defaultdict

from invenio_access.permissions import Identity, system_identity
from invenio_i18n import lazy_gettext as _
from invenio_pidstore.errors import PIDDoesNotExistError
from invenio_search.engine import dsl
from invenio_vocabularies.proxies import current_service as vocab_service

from invenio_communities.communities.records.systemfields.deletion_status import (
    CommunityDeletionStatusEnum,
)
from invenio_communities.proxies import current_communities

from .tasks import delete_community, restore_community


def _get_communities_for_user(user_id):
    """Helper function for getting all communities with the user as the sole owner.

    Note: This function performs DB queries yielding all communities with the given
    user as the sole owner (which is not hard-limited in the system) and performs
    service calls on each of them. Thus, this function has the potential of being a very
    heavy operation, and should not be called as part of the handling of an
    HTTP request!
    """
    comm_cls = current_communities.service.record_cls
    comm_model_cls = comm_cls.model_cls
    mem_cls = current_communities.service.members.record_cls
    mem_model_cls = mem_cls.model_cls

    # collect the owners for each community
    comm_owners = defaultdict(list)
    for comm_owner in [
        mem_cls(m.data, model=m)
        for m in mem_model_cls.query.filter(mem_model_cls.role == "owner").all()
    ]:
        comm_owners[comm_owner.community_id].append(comm_owner)

    # filter for communities that are owned solely by the user in question
    relevant_comm_ids = [
        comm_id
        for comm_id, owners in comm_owners.items()
        if len(owners) == 1 and str(owners[0].user_id) == user_id
    ]

    # resolve the communities in question
    communities = [
        comm_cls(m.data, model=m)
        for m in comm_model_cls.query.filter(
            comm_model_cls.id.in_(relevant_comm_ids)
        ).all()
    ]

    return communities


def on_block(user_id, uow=None, **kwargs):
    """Removes records that belong to a user.

    Note: This function operates on all records of a user and thus has the potential
    to be a very heavy operation! Thus it should not be called as part of the handling
    of an HTTP request!
    """
    user_id = str(user_id)
    tombstone_data = {"note": _("User was blocked")}

    # set the removal reason if the vocabulary item exists
    try:
        removal_reason_id = kwargs.get("removal_reason_id", "misconduct")
        vocab = vocab_service.read(
            identity=system_identity, id_=("removalreasons", removal_reason_id)
        )
        tombstone_data["removal_reason"] = {"id": vocab.id}
    except PIDDoesNotExistError:
        pass

    # soft-delete all the communities of that user (only if they are the only owner)
    for comm in _get_communities_for_user(user_id):
        delete_community.delay(comm.pid.pid_value, tombstone_data)


def on_restore(user_id, uow=None, **kwargs):
    """Restores records that belong to a user.

    Note: This function operates on all records of a user and thus has the potential
    to be a very heavy operation! Thus it should not be called as part of the handling
    of an HTTP request!
    """
    user_id = str(user_id)

    # restore all the deleted records of that user
    for comm in _get_communities_for_user(user_id):
        restore_community.delay(comm.pid.pid_value)


def on_approve(user_id, uow=None, **kwargs):
    """Execute on user approve.

    Re-index user records and dump verified field into records.
    """
    # import here due to circular import
    from invenio_communities.members.records.api import Member

    user_owned_communities = [
        m[0] for m in Member.get_memberships(Identity(user_id)) if m[1] == "owner"
    ]
    communities_filter = dsl.Q(
        "terms", **{"id": [id_ for id_ in user_owned_communities]}
    )
    community_service = current_communities.service
    community_service.reindex(system_identity, extra_filter=communities_filter)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2024 Northwestern University.
# Copyright (C) 2022-2023 Graz University of Technology.
#
# Invenio-Communities is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

from unittest.mock import MagicMock

from invenio_access.permissions import system_identity
from invenio_notifications.proxies import current_notifications_manager

from invenio_communities.notifications.builders import (
    CommunityInvitationAcceptNotificationBuilder,
    CommunityInvitationCancelNotificationBuilder,
    CommunityInvitationDeclineNotificationBuilder,
    CommunityInvitationExpireNotificationBuilder,
    CommunityInvitationSubmittedNotificationBuilder,
)


#
# invenio-notification testcases
#
def test_community_invitation_submit_notification(
    member_service,
    requests_service,
    community,
    owner,
    new_user,
    db,
    monkeypatch,
    app,
    clean_index,
):
    """Test notifcation being built on community invitation submit."""

    original_builder = CommunityInvitationSubmittedNotificationBuilder

    # mock build to observe calls
    mock_build = MagicMock()
    mock_build.side_effect = original_builder.build
    monkeypatch.setattr(original_builder, "build", mock_build)
    # setting specific builder for test case
    monkeypatch.setattr(
        current_notifications_manager,
        "builders",
        {
            **current_notifications_manager.builders,
            original_builder.type: original_builder,
        },
    )
    assert not mock_build.called

    mail = app.extensions.get("mail")
    assert mail

    with mail.record_messages() as outbox:
        # Validate that email was sent
        role = "reader"
        message = "<p>invitation message</p>"

        data = {
            "members": [{"type": "user", "id": str(new_user.id)}],
            "role": role,
            "message": message,
        }
        member_service.invite(owner.identity, community.id, data)
        # ensure that the invited user request has been indexed
        res = member_service.search_invitations(owner.identity, community.id).to_dict()
        assert res["hits"]["total"] == 1
        inv = res["hits"]["hits"][0]

        # check notification is build on submit
        assert mock_build.called
        assert len(outbox) == 1
        html = outbox[0].html
        # TODO: update to `req["links"]["self_html"]` when addressing https://github.com/inveniosoftware/invenio-rdm-records/issues/1327
        assert "/me/requests/{}".format(inv["request"]["id"]) in html
        # role titles will be capitalized
        assert role.capitalize() in html
        assert "You have been invited to join" in html
        assert message in html
        assert community["metadata"]["title"] in html

    # decline to reset
    requests_service.execute_action(new_user.identity, inv["request"]["id"], "decline")
    with mail.record_messages() as outbox:
        data = {
            "members": [{"type": "user", "id": str(new_user.id)}],
            "role": role,
        }
        # invite again without message
        member_service.invite(owner.identity, community.id, data)
        # ensure that the invited user request has been indexed
        res = member_service.search_invitations(owner.identity, community.id).to_dict()
        assert res["hits"]["total"] == 2
        inv = res["hits"]["hits"][1]

        # check notification is build on submit
        assert mock_build.called
        assert len(outbox) == 1
        html = outbox[0].html
        # TODO: update to `req["links"]["self_html"]` when addressing https://github.com/inveniosoftware/invenio-rdm-records/issues/1327
        assert "/me/requests/{}".format(inv["request"]["id"]) in html
        # role titles will be capitalized
        assert role.capitalize() in html
        assert "You have been invited to join" in html
        assert "with the following message:" not in html
        assert community["metadata"]["title"] in html


def test_community_invitation_accept_notification(
    member_service,
    requests_service,
    community,
    new_user,
    db,
    monkeypatch,
    app,
    members,
    clean_index,
):
    """Test notifcation sent on community invitation accept."""

    original_builder = CommunityInvitationAcceptNotificationBuilder

    owner = members["owner"]
    # mock build to observe calls
    mock_build = MagicMock()
    mock_build.side_effect = original_builder.build
    monkeypatch.setattr(original_builder, "build", mock_build)
    assert not mock_build.called

    mail = app.extensions.get("mail")
    assert mail

    role = "reader"
    data = {
        "members": [{"type": "user", "id": str(new_user.id)}],
        "role": role,
    }
    member_service.invite(owner.identity, community.id, data)
    res = member_service.search_invitations(owner.identity, community.id).to_dict()
    assert res["hits"]["total"] == 1
    inv = res["hits"]["hits"][0]
    with mail.record_messages() as outbox:
        # Validate that email was sent
        requests_service.execute_action(
            new_user.identity, inv["request"]["id"], "accept"
        )
        # check notification is build on submit
        assert mock_build.called
        # community owner, manager get notified
        assert len(outbox) == 2
        html = outbox[0].html
        # TODO: update to `req["links"]["self_html"]` when addressing https://github.com/inveniosoftware/invenio-rdm-records/issues/1327
        assert "/me/requests/{}".format(inv["request"]["id"]) in html
        # role titles will be capitalized
        assert (
            "'@{who}' accepted the invitation to join your community '{title}'".format(
                who=new_user.user.username
                or new_user.user.user_profile.get("full_name"),
                title=community["metadata"]["title"],
            )
            in html
        )


def test_community_invitation_cancel_notification(
    member_service,
    requests_service,
    community,
    owner,
    new_user,
    db,
    monkeypatch,
    app,
    clean_index,
):
    """Test notifcation sent on community invitation cancel."""

    original_builder = CommunityInvitationCancelNotificationBuilder

    # mock build to observe calls
    mock_build = MagicMock()
    mock_build.side_effect = original_builder.build
    monkeypatch.setattr(original_builder, "build", mock_build)
    assert not mock_build.called

    mail = app.extensions.get("mail")
    assert mail

    role = "reader"
    data = {
        "members": [{"type": "user", "id": str(new_user.id)}],
        "role": role,
    }

    member_service.invite(owner.identity, community.id, data)
    res = member_service.search_invitations(owner.identity, community.id).to_dict()
    assert res["hits"]["total"] == 1
    inv = res["hits"]["hits"][0]
    with mail.record_messages() as outbox:
        # Validate that email was sent
        requests_service.execute_action(owner.identity, inv["request"]["id"], "cancel")
        # check notification is build on submit
        assert mock_build.called
        # invited user gets notified
        assert len(outbox) == 1
        html = outbox[0].html
        # TODO: update to `req["links"]["self_html"]` when addressing https://github.com/inveniosoftware/invenio-rdm-records/issues/1327
        assert "/me/requests/{}".format(inv["request"]["id"]) in html
        # role titles will be capitalized
        assert (
            "The invitation for '@{who}' to join community '{title}' was cancelled".format(
                who=new_user.user.username
                or new_user.user.user_profile.get("full_name"),
                title=community["metadata"]["title"],
            )
            in html
        )


def test_community_invitation_decline_notification(
    member_service,
    requests_service,
    community,
    new_user,
    db,
    monkeypatch,
    app,
    members,
    clean_index,
):
    """Test notifcation sent on community invitation decline."""

    owner = members["owner"]
    original_builder = CommunityInvitationDeclineNotificationBuilder

    # mock build to observe calls
    mock_build = MagicMock()
    mock_build.side_effect = original_builder.build
    monkeypatch.setattr(original_builder, "build", mock_build)
    assert not mock_build.called

    mail = app.extensions.get("mail")
    assert mail

    role = "reader"
    data = {
        "members": [{"type": "user", "id": str(new_user.id)}],
        "role": role,
    }
    member_service.invite(owner.identity, community.id, data)
    res = member_service.search_invitations(owner.identity, community.id).to_dict()
    assert res["hits"]["total"] == 1
    inv = res["hits"]["hits"][0]
    with mail.record_messages() as outbox:
        # Validate that email was sent
        # Added resp
        resp = requests_service.execute_action(
            new_user.identity, inv["request"]["id"], "decline"
        )
        # check notification is build on submit
        assert mock_build.called
        # community owner, manager get notified
        assert len(outbox) == 2
        html = outbox[0].html
        # TODO: update to `req["links"]["self_html"]` when addressing https://github.com/inveniosoftware/invenio-rdm-records/issues/1327
        assert "/me/requests/{}".format(inv["request"]["id"]) in html
        # role titles will be capitalized
        assert (
            "'@{who}' declined the invitation to join your community '{title}'".format(
                who=new_user.user.username
                or new_user.user.user_profile.get("full_name"),
                title=community["metadata"]["title"],
            )
            in html
        )


def test_community_invitation_expire_notification(
    member_service,
    requests_service,
    community,
    new_user,
    db,
    monkeypatch,
    app,
    members,
    clean_index,
):
    """Test notifcation sent on community invitation decline."""

    owner = members["owner"]
    original_builder = CommunityInvitationExpireNotificationBuilder

    # mock build to observe calls
    mock_build = MagicMock()
    mock_build.side_effect = original_builder.build
    monkeypatch.setattr(original_builder, "build", mock_build)
    assert not mock_build.called

    mail = app.extensions.get("mail")
    assert mail

    role = "reader"
    data = {
        "members": [{"type": "user", "id": str(new_user.id)}],
        "role": role,
    }
    member_service.invite(owner.identity, community.id, data)
    res = member_service.search_invitations(owner.identity, community.id).to_dict()
    assert res["hits"]["total"] == 1
    inv = res["hits"]["hits"][0]
    with mail.record_messages() as outbox:
        # Validate that email was sent
        requests_service.execute_action(system_identity, inv["request"]["id"], "expire")

        # check notification is build on submit
        assert mock_build.called
        # community owner, manager and invited user get notified
        # TODO: Replace with equivalent
        assert len(outbox) == 3
        html = outbox[0].html
        # TODO: update to `req["links"]["self_html"]` when addressing https://github.com/inveniosoftware/invenio-rdm-records/issues/1327
        assert "/me/requests/{}".format(inv["request"]["id"]) in html
        # role titles will be capitalized
        assert (
            "The invitation for '@{who}' to join community '{title}' has expired.".format(
                who=new_user.user.username
                or new_user.user.user_profile.get("full_name"),
                title=community["metadata"]["title"],
            )
            in html
        )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022-2024 CERN.
#
# Invenio-Communities is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Utility for rendering URI template links."""

from invenio_records_resources.services.base.links import Link, LinksTemplate


class CommunityLinksTemplate(LinksTemplate):
    """Templates for generating links for a community object."""

    def __init__(self, links, action_link, available_actions, context=None):
        """Constructor."""
        super().__init__(links, context=context)
        self._action_link = action_link
        self._available_actions = available_actions

    def expand(self, identity, community):
        """Expand all the link templates."""
        links = {}

        # expand links for all available actions on the request
        link = self._action_link
        for action in self._available_actions:
            ctx = self.context.copy()
            ctx["action_name"] = action["action_name"]
            ctx["action"] = action["action_permission"]
            ctx["identity"] = identity
            if link.should_render(community, ctx):
                links[action["action_name"]] = link.expand(community, ctx)

        # expand the other configured links
        ctx = self.context.copy()
        ctx["identity"] = identity
        for key, link in self._links.items():
            if link.should_render(community, ctx):
                links[key] = link.expand(community, ctx)

        return links


class CommunityLink(Link):
    """Link variables setter for Community Members links."""

    @staticmethod
    def vars(record, vars):
        """Variables for the URI template."""
        vars.update(
            {
                "id": record.id,
                "slug": record.slug,
            }
        )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-Communities is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Schemas for parameter parsing."""

from invenio_records_resources.resources.records.args import SearchRequestArgsSchema
from marshmallow import fields


class CommunitiesSearchRequestArgsSchema(SearchRequestArgsSchema):
    """Extend schema with CSL fields."""

    status = fields.Str()
    include_deleted = fields.Bool()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2024 CERN.
#
# Invenio-Communities is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.
"""Configurations for subcommunities service."""

from invenio_records_permissions.generators import Disable, SystemProcess
from invenio_records_permissions.policies import BasePermissionPolicy
from invenio_records_resources.services.base.config import (
    ConfiguratorMixin,
    FromConfig,
    ServiceConfig,
)
from invenio_records_resources.services.records.results import RecordItem

from invenio_communities.generators import CommunityOwners
from invenio_communities.subcommunities.services.request import SubCommunityRequest

from .schema import SubcommunityRequestSchema


class SubCommunityPermissionPolicy(BasePermissionPolicy):
    """Configure permission policy for subcommunities."""

    can_request_join = [CommunityOwners()]
    can_read = [SystemProcess()]

    # Not supported actions
    can_create = [Disable()]
    can_search = [Disable()]
    can_update = [Disable()]
    can_delete = [Disable()]


class SubCommunityServiceConfig(ServiceConfig, ConfiguratorMixin):
    """SubCommunity service configuration."""

    service_id = "subcommunities"
    permission_policy_cls = SubCommunityPermissionPolicy
    result_item_cls = RecordItem
    result_list_cls = None  # TODO

    schema = FromConfig("COMMUNITIES_SUB_SERVICE_SCHEMA", SubcommunityRequestSchema)

    request_cls = FromConfig("COMMUNITIES_SUB_REQUEST_CLS", SubCommunityRequest)

    # links configuration
    links_item = {
        # TODO: add links
    }
